/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.3.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2009 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: dotk.inl,v 1.5 2009/08/20 08:48:08 cxsc Exp $ */

#ifndef _CXSC_DOTK_INL_INCLUDED
#define _CXSC_DOTK_INL_INCLUDED

#include "real.hpp"
#include <fenv.h>


namespace cxsc {
	
	/*
	   Sets the rounding mode according to the parameter r:
	   r=-1: Round downwards
           r=0 : Round to nearest
           r=1 : Round upwards
	   r=2 : Round toward zero
	*/
	static inline void setround(int r) {
		switch(r) {
			case -1:
				fesetround(FE_DOWNWARD);
				break;
			case 0:
				fesetround(FE_TONEAREST);
				break;
			case 1:
				fesetround(FE_UPWARD);
				break;
			case 2: 
				fesetround(FE_TOWARDZERO);
				break;
			default:
				fesetround(FE_TONEAREST);
		}
	}

	/*
	   Return an int value corresponding to the current rounding mode:
	   -1: Round downwards
           0 : Round to nearest
           1 : Round upwards
	   2 : Round toward zero
	*/
	static inline int getround() {
		switch(fegetround()) {
			case FE_DOWNWARD:
				return -1;
				break;
			case FE_TONEAREST:
				return 0;
				break;
			case FE_UPWARD:
				return 1;
				break;
			case FE_TOWARDZERO:
				return 2;
				break;
			default:
				return 0;
		}	
	}
	
	/*
	Converts the floating point sum of two floating point variables x and y into 
	a new sum of two floating point values a and b (x+y=a+b in the set of real 
	numbers), where a is the result of the floating point summation and b is a
	correction term ("the error of the floating point result")
	
	\param x First addend
	\param y Second addend
	\param a The result of the floating point summation
	\param b The error of the floating point summation
	*/ 
	static inline void TwoSum(const real x, const real &y, real &a, real &b) {
		a = x + y;
		real c = a - x;
		b = ((x - (a - c)) + (y - c));
	}
	
	/*
	Splits a floating point number x into two floating point number a and b
	with x=a+b, so that the mantissa of a and b are not overlapping and |y|<=|x|
	
	\param x Floating point number to be split
	\param x_h First part of x
	\param x_t Second part of x
	*/
	static inline void Split(const real &x, real &x_h, real &x_t) {
		real c = Factor * x;
		x_h = c - (c - x);
		x_t = x - x_h;
	}
	
	/*
	Computes the product of two floating point numbers as well as the error
	occuring during multiplication. A product x*y of two floating point numbers
	is transformed into the sum a+b of two other floating point numbers, where
	a is the result of the ordinary floating point multiplication and b
	is a correction value depicting the error.
	
	\param x First factor
	\param y Second factor
	\param a Result of the floating point multilpication
	\param b Error of the floating point multiplication
	*/
	static inline void TwoProduct(const real &x, const real &y, real &a, real &b) {
		real x1,x2,y1,y2;
		a = x * y;
		Split(x,x1,x2);
		Split(y,y1,y2);
		b = x2 * y2 - (((a - x1 * y1) - x2 * y1) - x1 * y2);
	} 

	/*
	Processes the error values according to the desired precision and 
	computes the final result of the DotK algorithm for K>=3 as well as an error
	bound.
	
	\param p Array consisting of correction values and floating point result of
		the dot product
	\param n Dimension of vector
	\param k Runs of SumK algorithm
	\param err Error bound
	\param val The final result of the Dotk-Algorithm stored in an accumulator
	*/
	static inline void SumK(real* p, int n, int k, real &err, dotprecision& val) {
		real corr = 0.0;
		val += p[n-1];
		
		for(int j=1 ; j<k ; j++) {
			for(int i=1 ; i<n-1 ; i++) 
				TwoSum(p[i],p[i-1],p[i],p[i-1]);    

			val += p[n-2];
			p[n-2] = 0.0;
		}
		
		for(int i=0 ; i<n-2 ; i++) {
			corr += p[i];
		}
	
		setround(1);
	
		for(int i=0 ; i<n-2 ; i++) {
			err += p[i];
		}
	
		err = err-corr;

		setround(0);
		
		val += corr;
	}


	/*
	Processes the error values according to the desired precision and 
	computes the final result of the DotK algorithm for K>=3. This version does
	NOT compute an error bound.
	
	\param p Array consisting of correction values and floating point result of
		the dot product
	\param n Dimension of vector
	\param k Runs of SumK algorithm
	\param val The final result of the Dotk-Algorithm stored in an accumulator
	*/
	template<typename T>
	static inline void SumK_NoErr(real* p, int n, int k, T& val) {
		real corr = 0.0;
		val += p[n-1];
		
		for(int j=1 ; j<k ; j++) {
			for(int i=1 ; i<n-1 ; i++) 
				TwoSum(p[i],p[i-1],p[i],p[i-1]);    

			val += p[n-2];
			p[n-2] = 0.0;
		}
		
		for(int i=0 ; i<n-2 ; i++) {
			corr += p[i];
		}
		
		val += corr;
	}


	/*
	Computes the dot product of two real vectors in k-fold double precision.
        Algorithms used for values of K (k is member of dotpecision object):
        k=0: Use accumulator (maximum accuracy)
        k=1: Use floating point computations (error bound computed via rounding mode switch)
        k>=2: Use DotK-algorithm
        The final result with an error bound is stored in an accumlator.
	
	\param val Accumulator storing the final result
	\param x First vector of dot product
	\param y Second vector of dot product
	*/
	template<typename S, typename T>
	inline void addDot(dotprecision &val, const S &x, const T &y) {
		int n = Ub(x)-Lb(x)+1;
		int lb1 = Lb(x);
		int lb2 = Lb(y);
		real res,err=0.0;
		int rnd;

		//Check rounding mode
		if((rnd=getround()) != 0) {
			setround(0);
		}
	
		if(val.k == 0) { //use accumulator
	
			for(int i=1 ; i<=n ; i++)
				accumulate(val, x[i+lb1-1], y[i+lb2-1]);
	
		} else if(val.k == 1) { //use floating point
	
			real resd = 0.0, resu = 0.0;
	
			setround(-1);
			for(int i=1 ; i<=n ; i++)
				resd += x[i+lb1-1] * y[i+lb2-1];
	
			setround(1);
			for(int i=1 ; i<=n ; i++)
				resu += x[i+lb1-1] * y[i+lb2-1];
	
			setround(0);
			res = resd+(resu-resd)*0.5;

                        setround(1);
			val.err = val.err + (resu-res);

			val += res;
	
		} else if(val.k == 2) { //use DotK optimized for K=2
	
			real p, s, h, r, q, t;
		
			TwoProduct(x[lb1],y[lb2],p,s);
	
			err += abs(s);
	
			for(int i=2 ; i<=n ; i++) {
				TwoProduct(x[lb1+i-1],y[lb2+i-1],h,r);
				TwoSum(p,h,p,q);
				t = q + r;
				s = s + t;
				err += abs(t);
			}
	
			val += p;
			val += s;
			res = p+s;
	
			real alpha, delta, error;
	
			delta = (n*Epsilon) / (1.0-2*n*Epsilon);
			alpha = (Epsilon*abs(res)) + (delta*err+3*MinReal/Epsilon);
			error = alpha / (1.0 - 2*Epsilon);
	
			setround(1);

			val.err += error;
	
		} else { //use DotK
	
			real r = 0, h;
			real* t = new real[2*n];
		
			for(int i=1 ; i<=n ; i++) {
				TwoProduct(x[lb1+i-1], y[lb2+i-1], h, t[i-1]);
				TwoSum(r, h, r, t[n+i-2]);
			}
	
			t[2*n-1] = r;
			SumK(t, 2*n, val.k-1, err, val);
	
			setround(1);
			val.err += err;
	
			delete[] t;
	
		}

		//Reset rounding mode to former value
		setround(rnd);
	
	}


	/*
	Computes the dot product of two real vectors in k-fold double precision.
        Algorithms used for values of K (k is member of dotpecision object):
        k=0: Use accumulator (maximum accuracy)
        k=1: Use floating point computations 
        k>=2: Use DotK-algorithm
        This version does not compute an error bound and is used for computations
	in the operators.
	
	\param val Accumulator storing the final result
	\param x First vector of dot product
	\param y Second vector of dot product
	*/
	template<typename S, typename T>
	inline void addDot_op(dotprecision &val, const S &x, const T &y) {
		int n = Ub(x)-Lb(x)+1;
		int lb1 = Lb(x);
		int lb2 = Lb(y);
		real res = 0.0;
		int rnd;

		//Check rounding mode
		if((rnd=getround()) != 0) {
			setround(0);
		}

		if(val.k == 0) { //Use accumulator
	
			for(int i=1 ; i<=n ; i++)
				accumulate(val, x[i+lb1-1], y[i+lb2-1]);
	
		} else if(val.k == 1) { //Use floating point
	
			for(int i=1 ; i<=n ; i++)
				res += x[i+lb1-1] * y[i+lb2-1];
	
			val += res;
	
		} else if(val.k == 2) { //Use optimized DotK
	
			real p, s, h, r, q, t;
	
			TwoProduct(x[lb1],y[lb2],p,s);
		
			for(int i=2 ; i<=n ; i++) {
				TwoProduct(x[lb1+i-1],y[lb2+i-1],h,r);
				TwoSum(p,h,p,q);
				t = q + r;
				s = s + t;
			}
	
			val += p;
			val += s;
	
		} else { //Use DotK
	
			real r = 0, h;
			real* t = new real[2*n];
	
			for(int i=1 ; i<=n ; i++) {
				TwoProduct(x[lb1+i-1], y[lb2+i-1], h, t[i-1]);
				TwoSum(r, h, r, t[n+i-2]);
			}
	
			t[2*n-1] = r;
			SumK_NoErr(t, 2*n, val.k-1, val);
	
			delete[] t;
	
		}

		//Reset rounding mode to former value
		setround(rnd);
	
	}

}

#endif // _CXSC_DOTK_INL_INCLUDED

